-- Ngaro VM
-- Copyright (c) 2010, Charles Childers
--
-- LuaBit
-- Copyright (c) 2006 - 2007 hanzhao (abrash_han@hotmail.com)
-- ----------------------------------------------------------------------------
-- Due to inclusion of the LuaBit module, this code is distributed under the
-- MIT License.
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
-- ----------------------------------------------------------------------------
-- LuaBit (subset)
do
local function check_int(n)
  if (n - math.floor(n) > 0) then
    error("trying to use bitwise operation on non-integer!")
  end
end

local function to_bits(n)
  check_int(n)
  if (n < 0) then
    return to_bits(bit.bnot(math.abs(n)) + 1)
  end
  local tbl = {}
  local cnt = 1
  while (n > 0) do
    local last = math.mod(n,2)
    if (last == 1) then tbl[cnt] = 1 else tbl[cnt] = 0 end
    n = (n-last)/2
    cnt = cnt + 1
  end
  return tbl
end

local function tbl_to_number(tbl)
  local n = table.getn(tbl)
  local rslt = 0
  local power = 1
  for i = 1, n do
    rslt = rslt + tbl[i]*power
    power = power*2
  end
  return rslt
end

local function expand(tbl_m, tbl_n)
  local big = {}
  local small = {}
  if (table.getn(tbl_m) > table.getn(tbl_n)) then
    big = tbl_m
    small = tbl_n
  else
    big = tbl_n
    small = tbl_m
  end
  for i = table.getn(small) + 1, table.getn(big) do
    small[i] = 0
  end
end

local function bit_or(m, n)
  local tbl_m = to_bits(m)
  local tbl_n = to_bits(n)
  expand(tbl_m, tbl_n)
  local tbl = {}
  local rslt = math.max(table.getn(tbl_m), table.getn(tbl_n))
  for i = 1, rslt do
    if (tbl_m[i]== 0 and tbl_n[i] == 0) then tbl[i] = 0 else tbl[i] = 1 end
  end
  return tbl_to_number(tbl)
end

local function bit_and(m, n)
  local tbl_m = to_bits(m)
  local tbl_n = to_bits(n)
  expand(tbl_m, tbl_n)
  local tbl = {}
  local rslt = math.max(table.getn(tbl_m), table.getn(tbl_n))
  for i = 1, rslt do
    if (tbl_m[i]== 0 or tbl_n[i] == 0) then tbl[i] = 0 else tbl[i] = 1 end
  end
  return tbl_to_number(tbl)
end

local function bit_not(n)
  local tbl = to_bits(n)
  local size = math.max(table.getn(tbl), 32)
  for i = 1, size do
    if (tbl[i] == 1) then tbl[i] = 0 else tbl[i] = 1 end
  end
  return tbl_to_number(tbl)
end

local function bit_xor(m, n)
  local tbl_m = to_bits(m)
  local tbl_n = to_bits(n)
  expand(tbl_m, tbl_n)
  local tbl = {}
  local rslt = math.max(table.getn(tbl_m), table.getn(tbl_n))
  for i = 1, rslt do
    if (tbl_m[i] ~= tbl_n[i]) then tbl[i] = 1 else tbl[i] = 0 end
  end
  return tbl_to_number(tbl)
end

local function bit_rshift(n, bits)
  check_int(n)
  local high_bit = 0
  if (n < 0) then
    n = bit_not(math.abs(n)) + 1
    high_bit = 2147483648
  end
  for i=1, bits do
    n = n/2
    n = bit_or(math.floor(n), high_bit)
  end
  return math.floor(n)
end

local function bit_lshift(n, bits)
  check_int(n)
  if (n < 0) then n = bit_not(math.abs(n)) + 1 end
  for i=1, bits do n = n*2 end
  return bit_and(n, 4294967295)
end

bit = {
  bnot = bit_not,
  band = bit_and,
  bor  = bit_or,
  bxor = bit_xor,
  brshift = bit_rshift,
  blshift = bit_lshift,
  tobits = to_bits,
  tonumb = tbl_to_number
}
end
-- ----------------------------------------------------------------------------
-- Variables
ip = 0                -- instruction pointer
sp = 0                -- stack pointer
rp = 0                -- return pointer
stack = {}            -- data stack
address = {}          -- return stack
memory = {}           -- simulated ram
ports = {}            -- io ports
files = {}            -- file pointers
i  = 0                -- generic loop index
-- ----------------------------------------------------------------------------
-- Support Code
function fsize(filename)
  fh = assert(io.open(filename, "rb"))
  len = assert(fh:seek("end"))
  fh:close()
  return len
end

function decode(b1, b2, b3, b4)
  if not b4 then error("need four bytes to convert to int",2) end
  local n = b1 + b2*256 + b3*65536 + b4*16777216
  n = (n > 2147483647) and (n - 4294967296) or n
  return n
end
-- ----------------------------------------------------------------------------
-- simulated device handler
function handle_devices()
  if ports[0] == 1 then return end
  if ports[0] == 0 then
    ports[0] = 1
    if ports[1] == 1 then
      ports[1] = string.byte(io.read(1))
      if ports[1] == 13 then ports[1] = 10 end
    end
    if ports[2] == 1 then
      if stack[sp] < 0 then
        io.write(string.char(27) .. "[2J" .. string.char(27) .. "[1;1H")
      elseif stack[sp] > 0 and stack[sp] < 128 then
        io.write(string.char(stack[sp]))
      end
      sp = sp - 1
      ports[2] = 0
   end
   if ports[4] == -1 then
    slot = 0
    i = 1
    while i < 8 do
      if files[i] == 0 then
        slot = i
      end
      i = i + 1
     end
     mode = stack[sp]
     sp = sp - 1
     i = stack[sp]
     sp = sp - 1
     name = ""
     while memory[i] ~= 0 do
       name = name .. string.char(memory[i])
       i = i + 1
     end
     if mode == 0 then files[slot] = io.open(name, "r") end
     if mode == 1 then files[slot] = io.open(name, "w") end
     if mode == 2 then files[slot] = io.open(name, "a") end
     if mode == 3 then files[slot] = io.open(name, "r+") end
     ports[4] = slot
   end
   if ports[4] == -2 then
     slot = stack[sp]
     sp = sp - 1
     name = files[slot]:read(1):byte(1)
     ports[4] = name
   end
   if ports[4] == -3 then
     slot = stack[sp]
     sp = sp - 1
     files[slot]:write(string.char(stack[sp]))
     sp = sp - 1
     ports[4] = 1
   end
   if ports[4] == -4 then
     slot = stack[sp]
     sp = sp - 1
     files[slot]:close()
     files[slot] = 0
     ports[4] = 0
   end
   if ports[4] == -5 then
     slot = stack[sp] + 1
     sp = sp - 1
     ports[4] = files[slot]:seek()
   end
   if ports[4] == -6 then
     slot = stack[sp]
     sp = sp - 1
     at = stack[sp]
     sp = sp - 1
     ports[4] = files[slot]:seek("cur", at)
   end
   if ports[4] == -7 then
     slot = stack[sp]
     sp = sp - 1
     at = files[slot]:seek("cur", 0)
     ports[4] = files[slot]:seek("end")
     files[slot]:seek("set", at)
   end
   if ports[4] == -8 then
     i = stack[sp]
     sp = sp - 1
     name = ""
     while memory[i] ~= 0 do
       name = name .. string.char(memory[i])
       i = i + 1
     end
     os.remove(name)
     ports[4] = 0
   end
   if ports[5] == -1 then  -- image size
     ports[5] = 1000000
   end
   if ports[5] == -2 then  -- canvas exists?
     ports[5] = 0
   end
   if ports[5] == -3 then  -- canvas width
     ports[5] = 0
   end
   if ports[5] == -4 then  -- canvas height
     ports[5] = 0
   end
   if ports[5] == -5 then  -- stack depth
     ports[5] = sp
   end
   if ports[5] == -6 then  -- address stack depth
     ports[5] = rp
   end
   if ports[5] == -7 then  -- mouse exists?
     ports[5] = 0
   end
   if ports[5] == -8 then  -- time
     ports[5] = os.time()
   end
   if ports[5] == -9 then  -- shutdown vm
     ip = 1000000
     ports[5] = 0
   end
   if ports[5] == -10 then -- environment
     ports[5] = 0
   end
   if ports[5] == -11 then -- console height
     ports[5] = 0
   end
   if ports[5] == -12 then -- console width
     ports[5] = 0
   end
  end
end
-- ----------------------------------------------------------------------------
-- process opcodes
function process()
  local a = 0
  local b = 0
  local c = 0
  local d = 0
  local opcode = memory[ip]
  if     opcode ==  0 then     -- nop
  elseif opcode ==  1 then     -- lit
    sp = sp + 1
    ip = ip + 1
    stack[sp] = memory[ip]
  elseif opcode ==  2 then     -- dup
    sp = sp + 1
    stack[sp] = stack[sp - 1]
  elseif opcode ==  3 then     -- drop
    sp = sp - 1
  elseif opcode ==  4 then     -- swap
    a = stack[sp]
    stack[sp] = stack[sp - 1]
    stack[sp - 1] = a
  elseif opcode ==  5 then     -- push
    rp = rp + 1
    address[rp] = stack[sp]
    sp = sp - 1
  elseif opcode ==  6 then     -- pop
    sp = sp + 1
    stack[sp] = address[rp]
    rp = rp - 1
  elseif opcode ==  7 then     -- loop
    stack[sp] = stack[sp] - 1
    if stack[sp] ~= 0 and stack[sp] > -1 then
      ip = ip + 1
      ip = memory[ip] - 1
    else
      ip = ip + 1
      sp = sp - 1
    end
  elseif opcode ==  8 then     -- jump
    ip = ip + 1
    ip = memory[ip] - 1
    if memory[ip + 1] == 0 then ip = ip + 1 end
    if memory[ip + 1] == 0 then ip = ip + 1 end
  elseif opcode ==  9 then     -- return
    ip = address[rp]
    rp = rp - 1
  elseif opcode == 10 then     -- >= jump
    ip = ip + 1
    if stack[sp - 1] > stack[sp] then ip = memory[ip] - 1 end
    sp = sp - 2
  elseif opcode == 11 then     -- <= jump
    ip = ip + 1
    if stack[sp - 1] < stack[sp] then ip = memory[ip] - 1 end
    sp = sp - 2
  elseif opcode == 12 then     -- != jump
    ip = ip + 1
    if stack[sp - 1] ~= stack[sp] then ip = memory[ip] - 1 end
    sp = sp - 2
  elseif opcode == 13 then     -- == jump
    ip = ip + 1
    if stack[sp - 1] == stack[sp] then ip = memory[ip] - 1 end
    sp = sp - 2
  elseif opcode == 14 then     -- @
    stack[sp] = memory[stack[sp]]
  elseif opcode == 15 then     -- !
    memory[stack[sp]] = stack[sp - 1]
    sp = sp - 2
  elseif opcode == 16 then     -- +
    stack[sp - 1] = stack[sp] + stack[sp - 1]
    sp = sp - 1
  elseif opcode == 17 then     -- -
    stack[sp - 1] = stack[sp - 1] - stack[sp]
    sp = sp - 1
  elseif opcode == 18 then     -- *
    stack[sp - 1] = stack[sp - 1] * stack[sp]
    sp = sp - 1
  elseif opcode == 19 then     -- /mod
    a = stack[sp]
    b = stack[sp - 1]
    stack[sp] = math.floor(b / a)
    stack[sp - 1] = b % a
  elseif opcode == 20 then     -- and
    stack[sp - 1] = bit.band(stack[sp - 1], stack[sp])
    sp = sp - 1
  elseif opcode == 21 then     -- or
    stack[sp - 1] = bit.bor(stack[sp - 1], stack[sp])
    sp = sp - 1
  elseif opcode == 22 then     -- xor
    stack[sp - 1] = bit.bxor(stack[sp - 1], stack[sp])
    sp = sp - 1
  elseif opcode == 23 then     -- <<
    stack[sp - 1] = bit.blshift(stack[sp - 1], stack[sp])
    sp = sp - 1
  elseif opcode == 24 then     -- >>
    stack[sp - 1] = bit.brshift(stack[sp - 1], stack[sp])
    sp = sp - 1
  elseif opcode == 25 then     -- 0;
    if stack[sp] == 0 then
      sp = sp - 1
      ip = address[rp]
      rp = rp - 1
    end
  elseif opcode == 26 then     -- inc
    stack[sp] = stack[sp] + 1
  elseif opcode == 27 then     -- dec
    stack[sp] = stack[sp] - 1
  elseif opcode == 28 then     -- in
    stack[sp] = ports[stack[sp]]
  elseif opcode == 29 then     -- out
    ports[stack[sp]] = stack[sp - 1]
    sp = sp - 2
  elseif opcode == 30 then     -- wait
    handle_devices()
  else                         -- call (implicit)
    rp = rp + 1
    address[rp] = ip
    ip = memory[ip] - 1
    if memory[ip + 1] == 0 then ip = ip + 1 end
    if memory[ip + 1] == 0 then ip = ip + 1 end
  end
end
-- ----------------------------------------------------------------------------
-- Load retroImage into memory
-- \ cells = size of image file (in cells)
-- \ image = pointer to retroImage file
cells = fsize("retroImage") / 4
image = io.open("retroImage", "rb")
while i < cells do
  memory[i] = decode(image:read(4):byte(1,4))
  i = i + 1
end
while i < 1000000 do
  memory[i] = 0
  i = i + 1
end

i = 0
while i < 1024 do
  stack[i] = 0
  address[i] = 0
  ports[i] = 0
  i = i + 1
end
ports[0] = 1
-- ----------------------------------------------------------------------------
-- Process the image
while ip < 1000000 do
  process()
  ip = ip + 1
end
